! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                 One-dimensional partitioning                 !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

! use clever search algorithm in fd4_part_1d_binsrch and fd4_part_1d_parallel
#define CLEVER_SEARCH
! when using CLEVER_SEARCH: use the binary search method (not tested!)
!#define CLEVER_BINARY_SEARCH

!! One-dimensional partitioning of prefix-summed weight arrays in partitions with
!! (almost) equal load. Used for space-filling curve partitioning.
!!
!! @author Matthias Lieber
module fd4_part_1d_mod

  use fd4_globaldef_mod
  use fd4_mpi_mod
  use timing_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  ! public methods
  public :: fd4_part_1d_binsrch
  public :: fd4_part_1d_parallel
  public :: fd4_part_1d_simple_h1
  public :: fd4_part_1d_simple_h2
  
contains  

  !! Search the largest value which is not larger than $sumload in the array of prefix summed block weights
  !! $blkweights_ps starting at $pstart and return the index in $blkweights_ps.
  !! Returns a value >=$pstart and <=$nblocks.
  !!
  !! If compiled with $CLEVER_SEARCH, use a more sophisticated algorithm which starts searching at index $guess.
  !! Returns a new value for $guess for subsequent searches.
  !!
  !! Internal function, called by $fd4_part_1d_binsrch and $fd4_part_1d_parallel.
  !!
  function fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) result(pend)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(in)    :: pstart                   !! start index for search
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k), intent(inout) :: guess                    !! guess for result value for clever algorithm
    integer(i_k) :: pend
    
    integer(i_k) :: i

#ifdef CLEVER_SEARCH
#ifdef CLEVER_BINARY_SEARCH
#warning compiling untested binary 1D search
    !CAUTION: binary search not tested, seems to be slower in many cases even with (guess-pstart)~100
    ! try a clever guess and then do binary search
    integer(i_k) :: p0,p1,plen,j
    if( blkweights_ps(guess) > sumload ) then
      ! guessed too many blocks
      p0=pstart
      p1=guess
      plen=p1-p0
      !do while(plen>1)
      do j=2,plen
        if(plen<=1) exit
        i=p0+ISHFT(plen,-1)
        if( blkweights_ps(i) > sumload ) then
          p1=i
        else
          p0=i
        end if
        plen=p1-p0
      end do
      pend = p0
    else
      ! guessed too few blocks (or exactly the right number of blocks - still need to check this)
      pend = nblocks
      do i=guess+1,nblocks
        if( blkweights_ps(i) > sumload ) then
          pend=i-1
          exit
        end if
      end do
    end if
    ! guess number of blocks for next rank = number of blocks for this rank
    guess=MIN(pend+pend-pstart,nblocks)
#else
    ! try a clever guess and then search with single steps forward or backward
    if( blkweights_ps(guess) > sumload ) then
      ! guessed too many blocks
      pend = pstart
      do i=guess-1,pstart,-1
        if( blkweights_ps(i) <= sumload ) then
          pend=i
          exit
        end if
      end do
    else
      ! guessed too few blocks (or exactly the right number of blocks - still need to check this)
      pend = nblocks
      do i=guess+1,nblocks
        if( blkweights_ps(i) > sumload ) then
          pend=i-1
          exit
        end if
      end do
    end if
    ! guess number of blocks for next rank = number of blocks for this rank
    guess=MIN(pend+pend-pstart,nblocks)
#endif
#else
#warning compiling with slow linear 1D search
    pend = nblocks
    do i=pstart+1,nblocks
      if( blkweights_ps(i) > sumload ) then
        pend = i - 1
        exit
      end if
    end do
#endif  
  end function fd4_part_1d_scan

  !! Calculate a well-balanced partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! The algorithm is exact for $eps=0.0 and large $maxsteps.
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - binary search for a bottleneck value until error <= $eps
  !!  - found bottleneck value is used to create partitioning
  !!  - it may happen that the last ranks end up having no local blocks, this is no bug
  !!
  !! Caution: Be sure that $maxweight is really the max difference of subsequent values in $blkweights_ps
  !! and not the maximum weight in the block weights before prefix sum. Small rounding errors due
  !! to limited computer precision can lead to different values. With a too small maxweight the algorithm
  !! misses the exit condition and thus runs $maxsteps steps.
  !! If the block weights do not utilize the full real(8) precision, this problem should not appear.
  !!
  !! The algorithm corresponds mostly the method EBS (exact bisection) from:
  !!
  !! Pinar, A. and C. Aykanat: Fast optimal load balancing algorithms for 1D partitioning.
  !! Journal of Parallel and Distributed Computing, 64(8):974-996, 2004.
  !!
  subroutine fd4_part_1d_binsrch(nblocks, nprocs, blkweights_ps, maxweight, eps, maxsteps, partition, maxload, reqsteps)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)  :: maxweight                !! max weight in weight array
    real(r8k),    intent(in)  :: eps                      !! max. difference of the resulting bottleneck value from the minimal
    integer(i_k), intent(in)  :: maxsteps                 !! max. number of binary search steps, may stop earlier depending on eps
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller
    integer(i_k), intent(out) :: reqsteps                 !! required number of search steps to reach the accuracy specified by eps

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    integer(i_k) :: guess
    real(r8k) :: minp1load ! smallest realizable bottleneck value greater then current targetload

    reqsteps = maxsteps
    totalweight = blkweights_ps(nblocks)
    maxload = totalweight
    
    avload = totalweight / nprocs   ! average proc load = optimal balance
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = avload + maxweight     ! upper bound for max proc load, worst value
    !write(*,'(A,E12.4,E12.4,A,F7.4,F7.4)') 'lb/ub: ',lbload,ubload,' ~ ',avload/lbload,avload/ubload
    
    
    !! find good approximation of minimal max proc load ("bottleneck value")

    bottleneck = ubload
    do step=1,maxsteps

      targetload = lbload * 0.5_r8k + ubload * 0.5_r8k
      guess = nblocks / nprocs

      sumload = targetload
      pstart = 0
      maxload = 0
      minp1load = totalweight
      do crank=0,nprocs-2
        ! find partition for rank crank with load <= targetload
        pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
        maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
        if(pend==nblocks) exit
        minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
        pstart = pend
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
      
      maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)
 
      if(sumload<totalweight) then
        ! miss
        lbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
      else
        ! hit
        bottleneck = maxload
        ubload = maxload
      end if
      
      !write(*,'(I6,E32.24,F7.4)') step,targetload,avload/targetload

      ! exit the loop if specified accuracy has been reached
      if(ubload-lbload<=eps) then
        reqsteps = step
        exit
      end if
      
    end do ! steps
    
    !print *,ubload,lbload,ubload-lbload

    !! finally create partition with given maximum load ("bottleneck value")

    guess = nblocks / nprocs
    pstart = 0
    sumload = bottleneck
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= bottleneck
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
      pstart = pend
      ! set partition vector
      partition(crank+1) = pstart+1
      sumload = bottleneck + blkweights_ps(pstart)
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck

  end subroutine fd4_part_1d_binsrch


  !! Calculate a well-balanced partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! The algorithm is exact for $eps=0.0 and large $maxsteps.
  !!
  !! Algorithm:
  !!  - embarrassingly parallel
  !!  - every process tries a different bottleneck value interval
  !!    - first each process checks its lower bound for the bottleneck
  !!    - if this is successfull, further local checks are skipped
  !!    - else binary search is applied within the interval until error <= $eps
  !!    - $maxsteps=0 disables the search, so that each rank checks its lower bound only
  !!  - best value found is determined with MPI_Allreduce
  !!  - found bottleneck value is used to create partitioning
  !!  - it may happen that the last ranks end up having no local blocks, this is no bug
  !!
  !! Caution: Be sure that $maxweight is really the max difference of subsequent values in $blkweights_ps
  !! and not the maximum weight in the block weights before prefix sum. Small rounding errors due
  !! to limited computer precision can lead to different values. With a too small maxweight the algorithm
  !! misses the exit condition and thus runs $maxsteps steps.
  !! If the block weights do not utilize the full real(8) precision, this problem should not appear.
  !!
  !! The bisection algorithm is based on the method EBS (exact bisection) from:
  !!
  !! Pinar, A. and C. Aykanat: Fast optimal load balancing algorithms for 1D partitioning.
  !! Journal of Parallel and Distributed Computing, 64(8):974-996, 2004.
  !!
  subroutine fd4_part_1d_parallel(nblocks, nprocs, blkweights_ps, maxweight, eps, maxsteps, mpi, &
                                          partition, maxload, reqsteps, mpi_time, err)
    integer(i_k), intent(in)   :: nblocks                  !! number of blocks
    integer(i_k), intent(in)   :: nprocs                   !! number of processes
    real(r8k),    intent(in)   :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)   :: maxweight                !! max weight in weight array
    real(r8k),    intent(in)   :: eps                      !! max. difference of the resulting bottleneck value from the minimal
    integer(i_k), intent(in)   :: maxsteps                 !! max. number of binary search steps, may stop earlier depending on eps
    type(fd4_mpi),intent(in)   :: mpi                      !! domain's MPI context
    integer(i_k), intent(out)  :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out)  :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller
    integer(i_k), intent(out)  :: reqsteps                 !! required number of search steps to reach the accuracy specified by eps
    integer(i8k), intent(out)  :: mpi_time                 !! the time of this rank for MPI_Allreduce in us
    integer(i_k), intent(out)  :: err                      !! error status: 0...ok

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    real(r8k) :: mybottleneck, mylbload, myubload
    integer(i_k) :: guess, mysteps
    real(r8k) :: minp1load ! smallest realizable bottleneck value greater then current targetload
    integer(i8k) :: timing_us0
    
    err = 0
    
    totalweight = blkweights_ps(nblocks)
    
    avload = totalweight / nprocs   ! average proc load = optimal balance
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = avload + maxweight     ! upper bound for max proc load, worst value
    

    !! every process starts with different [lb,ub] and checks if a partition exists with
    !! a bottleneck value within this interval

    mylbload = lbload +  mpi%rank    * (ubload-lbload) / mpi%nprocs
    myubload = lbload + (mpi%rank+1) * (ubload-lbload) / mpi%nprocs
    !write(*,'(I6,A,E12.4,E12.4,A,F7.4,F7.4)') mpi%rank,'  lb/ub: ',mylbload,myubload,' ~ ',avload/mylbload,avload/myubload


    !! find good approximation of minimal max proc load ("bottleneck value")

    !! start with checking the local mylbload, if this is valid, we can skip further search
    targetload = mylbload
    guess = nblocks / nprocs
    sumload = targetload
    pstart = 0
    maxload = 0
    minp1load = totalweight
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= targetload
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
      maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      if(pend==nblocks) exit
      minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
      pstart = pend
      sumload = targetload + blkweights_ps(pstart)
    end do ! crank

    maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)

    if(sumload<totalweight) then
      ! miss
      mylbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
      mybottleneck = ubload
      mysteps = maxsteps
    else
      ! hit
      mybottleneck = maxload
      mysteps = 0
    end if

    ! skip further search if specified accuracy has been reached
    if(myubload-mylbload<=eps) then
      mysteps = 0
    end if
    
    reqsteps = mysteps + 1

    !! now go through binary search
    do step=1,mysteps

      targetload = mylbload * 0.5_r8k + myubload * 0.5_r8k
      guess = nblocks / nprocs
      
      sumload = targetload
      pstart = 0
      maxload = 0
      minp1load = totalweight
      do crank=0,nprocs-2
        ! find partition for rank crank with load <= targetload
        pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
        maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
        if(pend==nblocks) exit
        minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
        pstart = pend
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
 
      maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)

      if(sumload<totalweight) then
        ! miss
        mylbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
      else
        ! hit
        mybottleneck = maxload
        myubload = maxload
      end if
      
      !write(*,'(I6,I6,E32.24,F7.4,L4)') mpi%rank,step,targetload,avload/targetload,sumload>=totalweight
      
      ! exit the loop if specified accuracy has been reached
      if(myubload-mylbload<=eps) then
        reqsteps = step + 1
        exit
      end if
      
    end do ! binary search steps
    
    !! every process has now its best found bottleneck value in variable mybottleneck
    !! if a process found no working bottleneck, this variable is equal ubload
    !! so the min value of all mybottleneck variables is the best found value
    
    call gettime_ofday(timing_us0)
    call MPI_Allreduce(mybottleneck, bottleneck, 1, MPI_R8K, MPI_MIN, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allreduce failed')
    call gettime_ofday(mpi_time)
    mpi_time = mpi_time - timing_us0


    !if(mpi%rank==0) write(*,'(A,F7.4)') 'and the winner is:',avload/bottleneck

    !! finally create partition with given maximum load ("bottleneck value")

    guess = nblocks / nprocs
    pstart = 0
    sumload = bottleneck
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= bottleneck
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
      pstart = pend
      ! set partition vector
      partition(crank+1) = pstart+1
      sumload = bottleneck + blkweights_ps(pstart)
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck

  end subroutine fd4_part_1d_parallel


  !! Calculate a partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - very simple heuristic
  !!  - fast
  !!
  !! This is simply the method described as Heuristic 1 in:
  !!
  !! Miguet, S. und J.-M. Pierson: Heuristics for 1D rectilinear partitioning as a low cost
  !! and high quality answer to dynamic load balancing. in: High-Performance Computing and
  !! Networking, LNCS, vol. 1225, pages 550-564. Springer, 1997.
  !!
  subroutine fd4_part_1d_simple_h1(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: crank, pstart, pend, guess
    real(r8k) :: sumload, bottleneck

    maxload = 0.0_r8k
    guess = nblocks / nprocs
    pstart = 0  ! last block of prev partition
    bottleneck = blkweights_ps(nblocks) / nprocs
    sumload = bottleneck
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= bottleneck
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) ! last block of current partition
      !maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      pstart = pend
      ! set partition vector
      partition(crank+1) = pstart+1
      sumload = bottleneck + sumload
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck
    !maxload = MAX(maxload, blkweights_ps(partition(nprocs)-1) - blkweights_ps(partition(nprocs-1)-1))

  end subroutine fd4_part_1d_simple_h1


  !! Calculate a partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - very simple heuristic
  !!  - fast
  !!
  !! This is simply the method described as Heuristic 2 in:
  !!
  !! Miguet, S. und J.-M. Pierson: Heuristics for 1D rectilinear partitioning as a low cost
  !! and high quality answer to dynamic load balancing. in: High-Performance Computing and
  !! Networking, LNCS, vol. 1225, pages 550-564. Springer, 1997.
  !!
  subroutine fd4_part_1d_simple_h2(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: crank, pstart, pend, guess
    real(r8k) :: sumload, bottleneck

    maxload = 0.0_r8k
    guess = nblocks / nprocs
    pstart = 0  ! last block of prev partition
    bottleneck = blkweights_ps(nblocks) / nprocs
    sumload = bottleneck
    do crank=0,nprocs-2
      ! find partition for rank crank with load <= bottleneck
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) ! last block of current partition
      if(pend < nblocks) then
        if(blkweights_ps(pend+1) - sumload < sumload - blkweights_ps(pend)) pend = pend + 1
      end if
      !maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      pstart = pend
      ! set partition vector
      partition(crank+1) = pstart+1
      sumload = bottleneck + sumload
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck
    !maxload = MAX(maxload, blkweights_ps(partition(nprocs)-1) - blkweights_ps(partition(nprocs-1)-1))

  end subroutine fd4_part_1d_simple_h2
  

end module fd4_part_1d_mod
